# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._tetrachoric.

Tetrachoric correlation coefficient
"""

from math import cos, pi

from ._token_distance import _TokenDistance

__all__ = ['Tetrachoric']


class Tetrachoric(_TokenDistance):
    r"""Tetrachoric correlation coefficient.

    For two sets X and Y and a population N, the Tetrachoric correlation
    coefficient :cite:`Pearson:1900` is

        .. math::

            corr_{Tetrachoric}(X, Y) = \cos \Big(\frac{\pi
            \sqrt{|X \setminus Y| \cdot |Y \setminus X|}}
            {\sqrt{|X \cap Y| \cdot |(N \setminus X) \setminus Y|} +
            \sqrt{|X \setminus Y| \cdot |Y \setminus X|}}\Big)

    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    this is

        .. math::

            corr_{Tetrachoric} =
            \cos \frac{\pi\sqrt{bc}}{\sqrt{ad}+\sqrt{bc}}

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        alphabet=None,
        tokenizer=None,
        intersection_type='crisp',
        **kwargs
    ):
        """Initialize Tetrachoric instance.

        Parameters
        ----------
        alphabet : Counter, collection, int, or None
            This represents the alphabet of possible tokens.
            See :ref:`alphabet <alphabet>` description in
            :py:class:`_TokenDistance` for details.
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(Tetrachoric, self).__init__(
            alphabet=alphabet,
            tokenizer=tokenizer,
            intersection_type=intersection_type,
            **kwargs
        )

    def corr(self, src, tar):
        """Return the Tetrachoric correlation coefficient of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Tetrachoric correlation coefficient

        Examples
        --------
        >>> cmp = Tetrachoric()
        >>> cmp.corr('cat', 'hat')
        0.9885309061036239
        >>> cmp.corr('Niall', 'Neil')
        0.9678978997263907
        >>> cmp.corr('aluminum', 'Catalan')
        0.7853000893691571
        >>> cmp.corr('ATCG', 'TAGC')
        -1.0


        .. versionadded:: 0.4.0

        """
        self._tokenize(src, tar)

        rad = (
            self._intersection_card() * self._total_complement_card()
        ) ** 0.5
        rbc = (self._src_only_card() * self._tar_only_card()) ** 0.5

        if rbc:
            return cos(pi * rbc / (rad + rbc))
        return 1.0

    def sim(self, src, tar):
        """Return the Tetrachoric correlation coefficient of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Tetrachoric correlation coefficient

        Examples
        --------
        >>> cmp = Tetrachoric()
        >>> cmp.sim('cat', 'hat')
        0.994265453051812
        >>> cmp.sim('Niall', 'Neil')
        0.9839489498631954
        >>> cmp.sim('aluminum', 'Catalan')
        0.8926500446845785
        >>> cmp.sim('ATCG', 'TAGC')
        0.0


        .. versionadded:: 0.4.0

        """
        return (1.0 + self.corr(src, tar)) / 2.0


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._tokenize.

_Tokenizer base class
"""

from collections import Counter
from math import exp, log1p, log2

__all__ = ['_Tokenizer']


class _Tokenizer(object):
    """Abstract _Tokenizer class.

    .. versionadded:: 0.4.0
    """

    def __init__(self, scaler=None, *args, **kwargs):
        """Initialize Tokenizer.

        Parameters
        ----------
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - 'length' : Each token has weight equal to its length.
                - 'length-log' : Each token has weight equal to the log of its
                   length + 1.
                - 'length-exp' : Each token has weight equal to e raised to its
                   length.
                - 'entropy' : Weights are scaled to the (log_2) information
                  entropy of each key's frequency.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.


        .. versionadded:: 0.4.0

        """
        super(_Tokenizer, self).__init__()

        self._scaler = scaler
        self._tokens = Counter()
        self._string = ''
        self._ordered_tokens = []
        self._ordered_weights = []

    def tokenize(self, string=None):
        """Tokenize the term and store it.

        The tokenized term is stored as an ordered list and as a Counter
        object.

        Parameters
        ----------
        string : str or None
            The string to tokenize


        .. versionadded:: 0.4.0
        .. versionchanged:: 0.4.1
            Added 'length', 'entropy', and related scalers

        """
        if string is not None:
            self._string = string
            self._ordered_tokens = [self._string]
            self._ordered_weights = [1]

        if self._scaler in {'SSK', 'length', 'length-log', 'length-exp'}:
            self._tokens = Counter()
            if self._scaler[:6] == 'length':
                self._ordered_weights = [len(_) for _ in self._ordered_tokens]
                if self._scaler == 'length-log':
                    self._ordered_weights = [
                        log1p(_) for _ in self._ordered_weights
                    ]
                elif self._scaler == 'length-exp':
                    self._ordered_weights = [
                        exp(_) for _ in self._ordered_weights
                    ]
            for token, weight in zip(
                self._ordered_tokens, self._ordered_weights
            ):
                self._tokens[token] += weight
        elif self._scaler == 'entropy':
            counts = Counter(self._ordered_tokens)
            n = len(self._ordered_tokens)
            self._tokens = {
                key: -(val / n) * log2(val / n) for key, val in counts.items()
            }
            self._ordered_weights = [
                self._tokens[tok] / counts[tok] for tok in self._ordered_tokens
            ]
        else:
            self._tokens = Counter(self._ordered_tokens)

        return self

    def count(self):
        """Return token count.

        Returns
        -------
        int
            The total count of tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.count()
        1


        .. versionadded:: 0.4.0

        """
        return sum(self.get_counter().values())

    def count_unique(self):
        """Return the number of unique elements.

        Returns
        -------
        int
            The number of unique tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.count_unique()
        1


        .. versionadded:: 0.4.0

        """
        return len(self._tokens.values())

    def get_counter(self):
        """Return the tokens as a Counter object.

        Returns
        -------
        Counter
            The Counter of tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.get_counter()
        Counter({'term': 1})


        .. versionadded:: 0.4.0

        """
        if self._scaler == 'set':
            return Counter({key: 1 for key in self._tokens.keys()})
        elif callable(self._scaler):
            return Counter(
                {key: self._scaler(val) for key, val in self._tokens.items()}
            )
        else:
            return self._tokens

    def get_set(self):
        """Return the unique tokens as a set.

        Returns
        -------
        Counter
            The set of tokens

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.get_set()
        {'term'}


        .. versionadded:: 0.4.0

        """
        return set(self._tokens.keys())

    def get_list(self):
        """Return the tokens as an ordered list.

        Returns
        -------
        Counter
            The list of q-grams in the order they were added.

        Examples
        --------
        >>> tok = _Tokenizer().tokenize('term')
        >>> tok.get_list()
        ['term']


        .. versionadded:: 0.4.0

        """
        return self._ordered_tokens

    def __repr__(self):
        """Return representation of tokens object.

        .. versionadded:: 0.4.0

        """
        return self.__class__.__name__ + '{}'.format(str(self._tokens)[7:])

    def __and__(self, other):
        """Return intersection with other tokens.

        .. versionadded:: 0.4.0

        """
        return self.get_counter() & other.get_counter()

    def __add__(self, other):
        """Return union with other tokens.

        .. versionadded:: 0.4.0

        """
        return self.get_counter() + other.get_counter()

    def __sub__(self, other):
        """Return difference from other tokens.

        .. versionadded:: 0.4.0

        """
        return self.get_counter() - other.get_counter()


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._stiles.

Stiles similarity
"""

from math import copysign, log10

from ._token_distance import _TokenDistance

__all__ = ['Stiles']


class Stiles(_TokenDistance):
    r"""Stiles similarity.

    For two sets X and Y and a population N, Stiles similarity
    :cite:`Stiles:1961` is

        .. math::

            sim_{Stiles}(X, Y) = log_{10}
            \frac{|N| \Big(||X \cap Y| \cdot
            |N| -
            |X \setminus Y| \cdot |Y \setminus X|| -
            \frac{|N|}{2}\Big)^2}
            {|X \setminus Y| \cdot |Y \setminus X| \cdot
            (|N| - |X \setminus Y|) \cdot
            (|N| - |Y \setminus X|)}

    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    this is

        .. math::

            sim_{Stiles} =
            log_{10} \frac{n(|an-bc|-\frac{1}{2}n)^2}{bc(n-b)(n-c)}

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        alphabet=None,
        tokenizer=None,
        intersection_type='crisp',
        **kwargs
    ):
        """Initialize Stiles instance.

        Parameters
        ----------
        alphabet : Counter, collection, int, or None
            This represents the alphabet of possible tokens.
            See :ref:`alphabet <alphabet>` description in
            :py:class:`_TokenDistance` for details.
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(Stiles, self).__init__(
            alphabet=alphabet,
            tokenizer=tokenizer,
            intersection_type=intersection_type,
            **kwargs
        )

    def sim_score(self, src, tar):
        """Return the Stiles similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Stiles similarity

        Examples
        --------
        >>> cmp = Stiles()
        >>> cmp.sim_score('cat', 'hat')
        2.6436977886009236
        >>> cmp.sim_score('Niall', 'Neil')
        2.1622951406967723
        >>> cmp.sim_score('aluminum', 'Catalan')
        0.41925115106844024
        >>> cmp.sim_score('ATCG', 'TAGC')
        -0.8426334527850912


        .. versionadded:: 0.4.0

        """
        self._tokenize(src, tar)

        eps = 0.0000001
        a = max(self._intersection_card(), eps)
        b = max(self._src_only_card(), eps)
        c = max(self._tar_only_card(), eps)
        n = max(self._total_complement_card(), eps) + a + b + c

        anmbc = a * n - b * c

        return copysign(
            log10(
                n
                * max((abs(anmbc) - n / 2) ** 2, eps)
                / (b * (n - b) * c * (n - c))
            ),
            anmbc,
        )

    def corr(self, src, tar):
        """Return the Stiles correlation of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Stiles correlation

        Examples
        --------
        >>> cmp = Stiles()
        >>> cmp.corr('cat', 'hat')
        0.14701542182970487
        >>> cmp.corr('Niall', 'Neil')
        0.11767566062554877
        >>> cmp.corr('aluminum', 'Catalan')
        0.022355640924908403
        >>> cmp.corr('ATCG', 'TAGC')
        -0.046296656196428934


        .. versionadded:: 0.4.0

        """
        return self.sim_score(src, tar) / max(
            self.sim_score(src, src), self.sim_score(tar, tar)
        )

    def sim(self, src, tar):
        """Return the normalized Stiles similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Normalized Stiles similarity

        Examples
        --------
        >>> cmp = Stiles()
        >>> cmp.sim('cat', 'hat')
        0.5735077109148524
        >>> cmp.sim('Niall', 'Neil')
        0.5588378303127743
        >>> cmp.sim('aluminum', 'Catalan')
        0.5111778204624542
        >>> cmp.sim('ATCG', 'TAGC')
        0.4768516719017855


        .. versionadded:: 0.4.0

        """
        return (1.0 + self.corr(src, tar)) / 2.0


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

r"""abydos.stats.

The stats module defines functions for calculating various statistical data
about linguistic objects.

Functions are provided for calculating the following means:

    - arithmetic mean (:py:func:`.amean`)
    - geometric mean (:py:func:`.gmean`)
    - harmonic mean (:py:func:`.hmean`)
    - quadratic mean (:py:func:`.qmean`)
    - contraharmonic mean (:py:func:`.cmean`)
    - logarithmic mean (:py:func:`.lmean`)
    - identric (exponential) mean (:py:func:`.imean`)
    - Seiffert's mean (:py:func:`.seiffert_mean`)
    - Lehmer mean (:py:func:`.lehmer_mean`)
    - Heronian mean (:py:func:`.heronian_mean`)
    - Hölder (power/generalized) mean (:py:func:`.hoelder_mean`)
    - arithmetic-geometric mean (:py:func:`.agmean`)
    - geometric-harmonic mean (:py:func:`.ghmean`)
    - arithmetic-geometric-harmonic mean (:py:func:`.aghmean`)

And for calculating:

    - midrange (:py:func:`.midrange`)
    - median (:py:func:`.median`)
    - mode (:py:func:`.mode`)
    - variance (:py:func:`.var`)
    - standard deviation (:py:func:`.std`)

Some examples of the basic functions:

>>> nums = [16, 49, 55, 49, 6, 40, 23, 47, 29, 85, 76, 20]
>>> amean(nums)
41.25
>>> aghmean(nums)
32.42167170892585
>>> heronian_mean(nums)
37.931508950381925
>>> mode(nums)
49
>>> std(nums)
22.876935255113754



Two pairwise functions are provided:

    - mean pairwise similarity (:py:func:`.mean_pairwise_similarity`), which
      returns the mean similarity (using a supplied similarity function) among
      each item in a collection
    - pairwise similarity statistics
      (:py:func:`.pairwise_similarity_statistics`), which returns the max, min,
      mean, and standard deviation of pairwise similarities between two
      collections

The confusion table class (:py:class:`.ConfusionTable`) can be constructed in
a number of ways:

    - four values, representing true positives, true negatives,
      false positives, and false negatives, can be passed to the constructor
    - a list or tuple with four values, representing true positives,
      true negatives, false positives, and false negatives, can be passed to
      the constructor
    - a dict with keys 'tp', 'tn', 'fp', 'fn', each assigned to the values for
      true positives, true negatives, false positives, and false negatives
      can be passed to the constructor

The :py:class:`.ConfusionTable` class has methods:

    - :py:meth:`.to_tuple` extracts the :py:class:`.ConfusionTable` values as a
      tuple: (:math:`w`, :math:`x`, :math:`y`, :math:`z`)
    - :py:meth:`.to_dict` extracts the :py:class:`.ConfusionTable` values as a
      dict: {'tp'::math:`w`, 'tn'::math:`x`, 'fp'::math:`y`, 'fn'::math:`z`}
    - :py:meth:`.true_pos` returns the number of true positives
    - :py:meth:`.true_neg` returns the number of true negatives
    - :py:meth:`.false_pos` returns the number of false positives
    - :py:meth:`.false_neg` returns the number of false negatives
    - :py:meth:`.correct_pop` returns the correct population
    - :py:meth:`.error_pop` returns the error population
    - :py:meth:`.pred_pos_pop` returns the test positive population
    - :py:meth:`.pred_neg_pop` returns the test negative population
    - :py:meth:`.cond_pos_pop` returns the condition positive population
    - :py:meth:`.cond_neg_pop` returns the condition negative population
    - :py:meth:`.population` returns the total population
    - :py:meth:`.precision` returns the precision
    - :py:meth:`.precision_gain` returns the precision gain
    - :py:meth:`.recall` returns the recall
    - :py:meth:`.specificity` returns the specificity
    - :py:meth:`.npv` returns the negative predictive value
    - :py:meth:`.fallout` returns the fallout
    - :py:meth:`.fdr` returns the false discovery rate
    - :py:meth:`.accuracy` returns the accuracy
    - :py:meth:`.accuracy_gain` returns the accuracy gain
    - :py:meth:`.balanced_accuracy` returns the balanced accuracy
    - :py:meth:`.informedness` returns the informedness
    - :py:meth:`.markedness` returns the markedness
    - :py:meth:`.pr_amean` returns the arithmetic mean of precision & recall
    - :py:meth:`.pr_gmean` returns the geometric mean of precision & recall
    - :py:meth:`.pr_hmean` returns the harmonic mean of precision & recall
    - :py:meth:`.pr_qmean` returns the quadratic mean of precision & recall
    - :py:meth:`.pr_cmean` returns the contraharmonic mean of precision &
      recall
    - :py:meth:`.pr_lmean` returns the logarithmic mean of precision & recall
    - :py:meth:`.pr_imean` returns the identric mean of precision & recall
    - :py:meth:`.pr_seiffert_mean` returns Seiffert's mean of precision &
      recall
    - :py:meth:`.pr_lehmer_mean` returns the Lehmer mean of precision & recall
    - :py:meth:`.pr_heronian_mean` returns the Heronian mean of precision &
      recall
    - :py:meth:`.pr_hoelder_mean` returns the Hölder mean of precision & recall
    - :py:meth:`.pr_agmean` returns the arithmetic-geometric mean of precision
      & recall
    - :py:meth:`.pr_ghmean` returns the geometric-harmonic mean of precision &
      recall
    - :py:meth:`.pr_aghmean` returns the arithmetic-geometric-harmonic mean of
      precision & recall
    - :py:meth:`.fbeta_score` returns the :math:`F_{beta}` score
    - :py:meth:`.f2_score` returns the :math:`F_2` score
    - :py:meth:`.fhalf_score` returns the :math:`F_{\frac{1}{2}}` score
    - :py:meth:`.e_score` returns the :math:`E` score
    - :py:meth:`.f1_score` returns the :math:`F_1` score
    - :py:meth:`.f_measure` returns the F measure
    - :py:meth:`.g_measure` returns the G measure
    - :py:meth:`.mcc` returns Matthews correlation coefficient
    - :py:meth:`.significance` returns the significance
    - :py:meth:`.kappa_statistic` returns the Kappa statistic


>>> ct = ConfusionTable(120, 60, 20, 30)
>>> ct.f1_score()
0.8275862068965518
>>> ct.mcc()
0.5367450401216932
>>> ct.specificity()
0.75
>>> ct.significance()
66.26190476190476


The :py:class:`.ConfusionTable` class also supports checking for equality with
another :py:class:`.ConfusionTable` and casting to string with ``str()``:

>>> (ConfusionTable({'tp':120, 'tn':60, 'fp':20, 'fn':30}) ==
... ConfusionTable(120, 60, 20, 30))
True
>>> str(ConfusionTable(120, 60, 20, 30))
'tp:120, tn:60, fp:20, fn:30'

----

"""

from ._confusion_table import ConfusionTable
from ._mean import (
    aghmean,
    agmean,
    amean,
    cmean,
    ghmean,
    gmean,
    heronian_mean,
    hmean,
    hoelder_mean,
    imean,
    lehmer_mean,
    lmean,
    median,
    midrange,
    mode,
    qmean,
    seiffert_mean,
    std,
    var,
)
from ._pairwise import mean_pairwise_similarity, pairwise_similarity_statistics

__all__ = [
    'ConfusionTable',
    'amean',
    'gmean',
    'hmean',
    'agmean',
    'ghmean',
    'aghmean',
    'cmean',
    'imean',
    'lmean',
    'qmean',
    'heronian_mean',
    'hoelder_mean',
    'lehmer_mean',
    'seiffert_mean',
    'median',
    'midrange',
    'mode',
    'std',
    'var',
    'mean_pairwise_similarity',
    'pairwise_similarity_statistics',
]


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.phonetic._phonic.

Phonic
"""

from ._phonetic import _Phonetic

__all__ = ['PHONIC']


class PHONIC(_Phonetic):
    """PHONIC code.

    PHONIC is a Soundex-like algorithm defined in :cite:`Taft:1970`.


    .. versionadded:: 0.4.1
    """

    _trans2 = {
        'CH': '6',
        'SH': '6',
        'PH': '8',
        'CE': '0',
        'CI': '0',
        'CY': '0',
    }

    _trans1 = {
        'D': '1',
        'T': '1',
        'N': '2',
        'M': '3',
        'R': '4',
        'L': '5',
        'J': '6',
        'C': '7',
        'K': '7',
        'G': '7',
        'Q': '7',
        'X': '7',
        'F': '8',
        'V': '8',
        'B': '9',
        'P': '9',
        'S': '0',
        'Z': '0',
    }

    _alphabetic = dict(zip((ord(_) for _ in '0123456789'), 'STNMRLJKFP'))

    def __init__(self, max_length=5, zero_pad=True, extended=False):
        """Initialize PHONIC instance.

        Parameters
        ----------
        max_length : int
            The length of the code returned (defaults to 5)
        zero_pad : bool
            Pad the end of the return value with 0s to achieve a max_length
            string
        extended : bool
            If True, this uses Taft's 'Extended PHONIC coding' mode, which
            simply omits the first character of the code.


        .. versionadded:: 0.4.1

        """
        # Require a max_length of at least 5 and not more than 64
        if max_length != -1:
            self._max_length = min(max(5, max_length), 64)
        else:
            self._max_length = 64

        self._zero_pad = zero_pad
        self._extended = extended

    def encode_alpha(self, word):
        """Return the alphabetic PHONIC code for a word.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        str
            The alphabetic PHONIC value

        Examples
        --------
        >>> pe = PHONIC()
        >>> pe.encode_alpha('Christopher')
        'JRSTF'
        >>> pe.encode_alpha('Niall')
        'NL'
        >>> pe.encode_alpha('Smith')
        'SMT'
        >>> pe.encode_alpha('Schmidt')
        'SJMT'


        .. versionadded:: 0.4.1

        """
        save_pad = self._zero_pad
        save_ext = self._extended
        self._zero_pad = False
        self._extended = True
        code = self.encode(word)
        self._zero_pad = save_pad
        self._extended = save_ext
        return code.translate(self._alphabetic)

    def encode(self, word):
        """Return the PHONIC code for a word.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        str
            The PHONIC code

        Examples
        --------
        >>> pe = PHONIC()
        >>> pe.encode('Christopher')
        'C6401'
        >>> pe.encode('Niall')
        'N2500'
        >>> pe.encode('Smith')
        'S0310'
        >>> pe.encode('Schmidt')
        'S0631'


        .. versionadded:: 0.4.1

        """
        # uppercase
        word = word.upper()

        code = []
        pos = 0
        while pos < len(word):
            if word[pos : pos + 2] in self._trans2:
                code.append(self._trans2[word[pos : pos + 2]])
                pos += 1
            elif word[pos] in self._trans1:
                code.append(self._trans1[word[pos]])
            else:
                code.append('.')
            pos += 1

        code = ''.join(code)
        code = self._delete_consecutive_repeats(code)
        code = code.replace('.', '')

        if self._zero_pad:
            code += '0' * (self._max_length - 1 - len(code))

        if not self._extended:
            code = word[:1] + code

        return code[: self._max_length]


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._dunning.

Dunning similarity
"""

from math import log

from ._token_distance import _TokenDistance

__all__ = ['Dunning']


class Dunning(_TokenDistance):
    r"""Dunning similarity.

    For two sets X and Y and a population N, Dunning log-likelihood
    :cite:`Dunning:1993`, following :cite:`Church:1991`, is

        .. math::

            sim_{Dunning}(X, Y) = \lambda =
            |X \cap Y| \cdot log_2(|X \cap Y|) +\\
            |X \setminus Y| \cdot log_2(|X \setminus Y|) +
            |Y \setminus X| \cdot log_2(|Y \setminus X|) +\\
            |(N \setminus X) \setminus Y| \cdot
            log_2(|(N \setminus X) \setminus Y|) -\\
            (|X| \cdot log_2(|X|) +
            |Y| \cdot log_2(|Y|) +\\
            |N \setminus Y| \cdot log_2(|N \setminus Y|) +
            |N \setminus X| \cdot log_2(|N \setminus X|))


    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    this is

        .. math::

            sim_{Dunning} = \lambda =
            a \cdot log_2(a) +\\
            b \cdot log_2(b) + c \cdot log_2(c) +
            d \cdot log_2(d) - \\
            ((a+b) \cdot log_2(a+b) + (a+c) \cdot log_2(a+c) +\\
            (b+d) \cdot log_2(b+d) + (c+d) log_2(c+d))

    Notes
    -----
    To avoid NaNs, every logarithm is calculated as the logarithm of 1 greater
    than the value in question. (Python's math.log1p function is used.)


    .. versionadded:: 0.4.0

    """

    def __init__(
        self,
        alphabet=None,
        tokenizer=None,
        intersection_type='crisp',
        **kwargs
    ):
        """Initialize Dunning instance.

        Parameters
        ----------
        alphabet : Counter, collection, int, or None
            This represents the alphabet of possible tokens.
            See :ref:`alphabet <alphabet>` description in
            :py:class:`_TokenDistance` for details.
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(Dunning, self).__init__(
            alphabet=alphabet,
            tokenizer=tokenizer,
            intersection_type=intersection_type,
            **kwargs
        )

    def sim_score(self, src, tar):
        """Return the Dunning similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Dunning similarity

        Examples
        --------
        >>> cmp = Dunning()
        >>> cmp.sim('cat', 'hat')
        0.33462839191969423
        >>> cmp.sim('Niall', 'Neil')
        0.19229445539929793
        >>> cmp.sim('aluminum', 'Catalan')
        0.03220862737070572
        >>> cmp.sim('ATCG', 'TAGC')
        0.0010606026735052122


        .. versionadded:: 0.4.0

        """
        self._tokenize(src, tar)

        a = self._intersection_card()
        b = self._src_only_card()
        c = self._tar_only_card()
        d = self._total_complement_card()
        n = a + b + c + d

        # a should not equal n, because 0 will result
        # As a workaround, we set d to 1 and add one to n.
        if a == n:
            d = 1
            n += 1

        a /= n
        b /= n
        c /= n
        d /= n

        score = 0.0
        for i in [a, b, c, d]:
            if i > 0:
                score += i * log(i)
        for i in [a, d]:
            for j in [b, c]:
                ij = i + j
                if ij > 0:
                    score -= ij * log(ij)
        score *= 2
        score /= log(2)

        return abs(round(score, 15))

    def sim(self, src, tar):
        """Return the normalized Dunning similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Normalized Dunning similarity

        Examples
        --------
        >>> cmp = Dunning()
        >>> cmp.sim('cat', 'hat')
        0.33462839191969423
        >>> cmp.sim('Niall', 'Neil')
        0.19229445539929793
        >>> cmp.sim('aluminum', 'Catalan')
        0.03220862737070572
        >>> cmp.sim('ATCG', 'TAGC')
        0.0010606026735052122


        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 1.0

        score = self.sim_score(src, tar)
        if not score:
            return 0.0

        norm = max(self.sim_score(src, src), self.sim_score(tar, tar))
        return score / norm


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.corpus._unigram_corpus.

Unigram Corpus
"""

import pickle  # noqa: S403
from codecs import open as c_open
from collections import Counter, defaultdict
from math import log1p

__all__ = ['UnigramCorpus']


def _dd_default(*args):
    return 0, 0


class UnigramCorpus(object):
    """Unigram corpus class.

    Largely intended for calculating inverse document frequence (IDF) from a
    large corpus of unigram (or smaller) tokens, this class encapsulates a
    dict object. Each key is a unigram token whose value is a tuple consisting
    of the number of times a term appeared and the number of distinct documents
    in which it appeared.

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        corpus_text='',
        documents=0,
        word_transform=None,
        word_tokenizer=None,
    ):
        r"""Initialize UnigramCorpus.

        Parameters
        ----------
        corpus_text : str
            The corpus text as a single string
        documents : int
            The number of documents in the corpus. If equal to 0 (the default)
            then the maximum from the internal dictionary's distinct
            documents count.
        word_transform : function
            A function to apply to each term before term tokenization and
            addition to the corpus. One might use this, for example, to apply
            Soundex encoding to each term.
        word_tokenizer : _Tokenizer
            A tokenizer to apply to each sentence in order to retrieve the
            individual "word" tokens. If set to none, str.split() will be used.

        Example
        -------
        >>> tqbf = 'The quick brown fox jumped over the lazy dog.\n'
        >>> tqbf += 'And then it slept.\n And the dog ran off.'
        >>> corp = UnigramCorpus(tqbf)


        .. versionadded:: 0.4.0

        """
        self.corpus = defaultdict(_dd_default)
        self.transform = word_transform
        self.tokenizer = word_tokenizer
        self.doc_count = documents

        self.add_document(corpus_text)

    def add_document(self, doc):
        """Add a new document to the corpus.

        Parameters
        ----------
        doc : str
            A string, representing the document to be added.


        .. versionadded:: 0.4.0

        """
        for word, count in Counter(doc.split()).items():
            self._add_word(word, count, 1)
        self.doc_count += 1

    def save_corpus(self, filename):
        """Save the corpus to a file.

        This employs pickle to save the corpus (a defaultdict). Other
        parameters of the corpus, such as its word_tokenizer, will not be
        affected and should be set during initialization.

        Parameters
        ----------
        filename : str
            The filename to save the corpus to.


        .. versionadded:: 0.4.0

        """
        with open(filename, mode='wb') as pkl:
            pickle.dump(self.corpus, pkl)

    def load_corpus(self, filename):
        """Load the corpus from a file.

        This employs pickle to load the corpus (a defaultdict). Other
        parameters of the corpus, such as its word_tokenizer, will not be
        affected and should be set during initialization.

        Parameters
        ----------
        filename : str
            The filename to load the corpus from.


        .. versionadded:: 0.4.0

        """
        with open(filename, mode='rb') as pkl:
            self.corpus = pickle.load(pkl)  # noqa: S301
        self._update_doc_count()

    def _update_doc_count(self):
        """Update document count, if necessary.

        .. versionadded:: 0.4.0
        """
        max_docs = max(self.corpus.values(), key=lambda _: _[1])[1]
        self.doc_count = max(max_docs, self.doc_count)

    def _add_word(self, word, count, doc_count):
        """Add a term to the corpus, possibly after tokenization.

        Parameters
        ----------
        word : str
            Word to add to the corpus
        count : int
            Count of word appearances
        doc_count : int
            Count of distinct documents in which word appears


        .. versionadded:: 0.4.0

        """
        if self.transform is not None:
            word = self.transform(word)

        if self.tokenizer is not None:
            tokens = self.tokenizer.tokenize(word).get_counter()
            for tok in tokens:
                n = tokens[tok] * count
                prior_count, prior_doc_count = self.corpus[tok]
                self.corpus[tok] = (
                    prior_count + n,
                    prior_doc_count + doc_count,
                )
        else:
            prior_count, prior_doc_count = self.corpus[word]
            self.corpus[word] = (
                prior_count + count,
                prior_doc_count + doc_count,
            )

    def gng_importer(self, corpus_file):
        """Fill in self.corpus from a Google NGram corpus file.

        Parameters
        ----------
        corpus_file : file
            The Google NGram file from which to initialize the n-gram corpus


        .. versionadded:: 0.4.0

        """
        with c_open(corpus_file, 'r', encoding='utf-8') as gng:
            for line in gng:
                line = line.rstrip().split('\t')
                word = line[0]
                if '_' in word:
                    word = word[: word.find('_')]

                self._add_word(word, int(line[2]), int(line[3]))
            self._update_doc_count()

    def idf(self, term):
        r"""Calculate the Inverse Document Frequency of a term in the corpus.

        Parameters
        ----------
        term : str
            The term to calculate the IDF of

        Returns
        -------
        float
            The IDF

        Examples
        --------
        >>> tqbf = 'the quick brown fox jumped over the lazy dog\n\n'
        >>> tqbf += 'and then it slept\n\n and the dog ran off'
        >>> corp = UnigramCorpus(tqbf)
        >>> round(corp.idf('dog'), 10)
        0.6931471806
        >>> round(corp.idf('the'), 10)
        0.6931471806


        .. versionadded:: 0.4.0

        """
        if term in self.corpus:
            count, term_doc_count = self.corpus[term]
            return log1p(self.doc_count / term_doc_count)
        else:
            return float('inf')


if __name__ == '__main__':
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.fingerprint._extract.

Taft's extract letter list coding
"""

from ._fingerprint import _Fingerprint

__all__ = ['Extract']


class Extract(_Fingerprint):
    """Extract Letter List fingerprint.

    Based on the extract letter list coding from :cite:`Taft:1970`,
    for lists 1, 2, 3, & 4.

    .. versionadded:: 0.4.1
    """

    def __init__(self, letter_list=1):
        """Initialize Extract instance.

        Parameters
        ----------
        letter_list : int or iterable
            If an integer (1-4) is supplied, Taft's specified letter lists are
            used. If an iterable is supplied, its values will be used as the
            list of letters to remove (in order).


        .. versionadded:: 0.4.1

        """
        letter_lists = [
            'ETAONRISHDLFCMUGYPWBVKXJQZ',
            'ETASIONRHCDLPMFBUWGYKVJQZX',
            'ETAONISRHLDCUMFYWGPKBVXJQZ',
            'EARNLOISTHDMCBGUWYJKPFVZXQ',
        ]

        super(_Fingerprint, self).__init__()
        self._letter_list = letter_list
        if isinstance(self._letter_list, int) and 1 <= self._letter_list <= 4:
            self._letter_list = list(letter_lists[self._letter_list - 1])
        elif hasattr(self._letter_list, '__iter__'):
            self._letter_list = list(self._letter_list)
        else:
            self._letter_list = list(letter_lists[0])

    def fingerprint(self, word):
        """Return the extract letter list coding.

        Parameters
        ----------
        word : str
            The word to fingerprint

        Returns
        -------
        int
            The extract letter list coding

        Examples
        --------
        >>> fp = Extract()
        >>> fp.fingerprint('hat')
        'HAT'
        >>> fp.fingerprint('niall')
        'NILL'
        >>> fp.fingerprint('colin')
        'CLIN'
        >>> fp.fingerprint('atcg')
        'ATCG'
        >>> fp.fingerprint('entreatment')
        'NRMN'


        .. versionadded:: 0.4.1

        """
        # uppercase & reverse
        word = word.upper()[::-1]

        for letter in self._letter_list:  # pragma: no branch
            if len(word) < 5:
                break

            count = word.count(letter)
            if count:
                word = word.replace(
                    letter, '', count - (4 - (len(word) - count))
                )

        return word[::-1]


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._higuera_mico.

The Higuera-Micó contextual normalized edit distance
"""

from numpy import full as np_full

from ._distance import _Distance

__all__ = ['HigueraMico']


class HigueraMico(_Distance):
    """The Higuera-Micó contextual normalized edit distance.

    This is presented in :cite:`Higuera:2008`.

    This measure is not normalized to a particular range. Indeed, for an
    string of infinite length as and a string of 0 length, the contextual
    normalized edit distance would be infinity. But so long as the relative
    difference in string lengths is not too great, the distance will generally
    remain below 1.0

    Notes
    -----
    The "normalized" version of this distance, implemented in the dist method
    is merely the minimum of the distance and 1.0.

    .. versionadded:: 0.4.0

    """

    def __init__(self, **kwargs):
        """Initialize Levenshtein instance.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        super(HigueraMico, self).__init__(**kwargs)

    def dist_abs(self, src, tar):
        """Return the Higuera-Micó distance between two strings.

        This is a straightforward implementation of Higuera & Micó pseudocode
        from :cite:`Higuera:2008`, ported to Numpy.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            The Higuera-Micó distance between src & tar

        Examples
        --------
        >>> cmp = HigueraMico()
        >>> cmp.dist_abs('cat', 'hat')
        0.3333333333333333
        >>> cmp.dist_abs('Niall', 'Neil')
        0.5333333333333333
        >>> cmp.dist_abs('aluminum', 'Catalan')
        0.7916666666666667
        >>> cmp.dist_abs('ATCG', 'TAGC')
        0.6000000000000001

        .. versionadded:: 0.4.0

        """
        if src == tar:
            return 0.0

        mx = np_full(
            (len(src) + 1, len(tar) + 1, len(src) + len(tar) + 1),
            fill_value=float('-inf'),
            dtype=float,
        )

        for i in range(1, len(src) + 1):
            mx[i, 0, i] = 0
        for j in range(len(tar) + 1):
            mx[0, j, j] = j
        for i in range(1, len(src) + 1):
            for j in range(1, len(tar) + 1):
                if src[i - 1] == tar[j - 1]:
                    for k in range(len(src) + len(tar) + 1):
                        mx[i, j, k] = mx[i - 1, j - 1, k]
                else:
                    for k in range(1, len(src) + len(tar) + 1):
                        mx[i, j, k] = mx[i - 1, j - 1, k - 1]
                for k in range(1, len(src) + len(tar) + 1):
                    mx[i, j, k] = max(
                        mx[i - 1, j, k - 1],
                        mx[i, j - 1, k - 1] + 1,
                        mx[i, j, k],
                    )

        min_dist = float('inf')
        for k in range(len(src) + len(tar) + 1):
            if mx[len(src), len(tar), k] >= 0:
                n_i = int(mx[len(src), len(tar), k])
                n_d = len(src) - len(tar) + n_i
                n_s = k - (n_i + n_d)
                loc_dist = 0
                for i in range(len(src) + 1, len(src) + n_i + 1):
                    loc_dist += 1 / i
                loc_dist += n_s / (len(src) + n_i)
                for i in range(len(tar) + 1, len(tar) + n_d + 1):
                    loc_dist += 1 / i
                if loc_dist < min_dist:
                    min_dist = loc_dist

        return min_dist

    def dist(self, src, tar):
        """Return the bounded Higuera-Micó distance between two strings.

        This is the distance bounded to the range [0, 1].

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            The bounded Higuera-Micó distance between src & tar

        Examples
        --------
        >>> cmp = HigueraMico()
        >>> cmp.dist('cat', 'hat')
        0.3333333333333333
        >>> cmp.dist('Niall', 'Neil')
        0.5333333333333333
        >>> cmp.dist('aluminum', 'Catalan')
        0.7916666666666667
        >>> cmp.dist('ATCG', 'TAGC')
        0.6000000000000001

        .. versionadded:: 0.4.0

        """
        return min(1.0, self.dist_abs(src, tar))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._goodman_kruskal_tau_a.

Goodman & Kruskal's Tau A similarity
"""

from ._token_distance import _TokenDistance

__all__ = ['GoodmanKruskalTauA']


class GoodmanKruskalTauA(_TokenDistance):
    r"""Goodman & Kruskal's Tau A similarity.

    For two sets X and Y and a population N, Goodman & Kruskal's :math:`\tau_a`
    similarity :cite:`Goodman:1954`, by analogy with :math:`\tau_b`, is

        .. math::

            sim_{GK_{\tau_a}}(X, Y) =
            \frac{\frac{\frac{|X \cap Y|}{|N|}^2 +
            \frac{|Y \setminus X|}{|N|}^2}{\frac{|Y|}{|N|}}+
            \frac{\frac{|X \setminus Y|}{|N|}^2 +
            \frac{|(N \setminus X) \setminus Y|}{|N|}^2}
            {\frac{|N \setminus X|}{|N|}} -
            (\frac{|X|}{|N|}^2 + \frac{|N \setminus X|}{|N|}^2)}
            {1 - (\frac{|X|}{|N|}^2 + \frac{|N \setminus X|}{|N|}^2)}

    In :ref:`2x2 confusion table terms <confusion_table>`, where a+b+c+d=n,
    after each term has been converted to a proportion by dividing by n, this
    is

        .. math::

            sim_{GK_{\tau_a}} =
            \frac{
            \frac{a^2 + c^2}{a+c} +
            \frac{b^2 + d^2}{b+d} -
            ((a+b)^2 + (c+d)^2)}
            {1 - ((a+b)^2 + (c+d)^2)}

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        alphabet=None,
        tokenizer=None,
        intersection_type='crisp',
        normalizer='proportional',
        **kwargs
    ):
        """Initialize GoodmanKruskalTauA instance.

        Parameters
        ----------
        alphabet : Counter, collection, int, or None
            This represents the alphabet of possible tokens.
            See :ref:`alphabet <alphabet>` description in
            :py:class:`_TokenDistance` for details.
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package
        intersection_type : str
            Specifies the intersection type, and set type as a result:
            See :ref:`intersection_type <intersection_type>` description in
            :py:class:`_TokenDistance` for details.
        normalizer : str
            Specifies the normalization type. See :ref:`normalizer <alphabet>`
            description in :py:class:`_TokenDistance` for details.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.
        metric : _Distance
            A string distance measure class for use in the ``soft`` and
            ``fuzzy`` variants.
        threshold : float
            A threshold value, similarities above which are counted as
            members of the intersection for the ``fuzzy`` variant.


        .. versionadded:: 0.4.0

        """
        super(GoodmanKruskalTauA, self).__init__(
            alphabet=alphabet,
            tokenizer=tokenizer,
            intersection_type=intersection_type,
            normalizer=normalizer,
            **kwargs
        )

    def sim(self, src, tar):
        """Return Goodman & Kruskal's Tau A similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            Goodman & Kruskal's Tau A similarity

        Examples
        --------
        >>> cmp = GoodmanKruskalTauA()
        >>> cmp.sim('cat', 'hat')
        0.3304969657208484
        >>> cmp.sim('Niall', 'Neil')
        0.22137604585914503
        >>> cmp.sim('aluminum', 'Catalan')
        0.05991264724130685
        >>> cmp.sim('ATCG', 'TAGC')
        4.119695274745721e-05


        .. versionadded:: 0.4.0

        """
        self._tokenize(src, tar)

        a = self._intersection_card()
        b = self._src_only_card()
        c = self._tar_only_card()
        d = self._total_complement_card()

        if a + b == 0 or a + c == 0:
            return 0.0

        fp = (a * a + c * c) / (a + c)

        sp = b * b + d * d
        if sp:
            sp /= b + d

        num = fp + sp - (a + b) ** 2 - (c + d) ** 2
        if num > 1e-14:
            return num / (1 - (a + b) ** 2 - (c + d) ** 2)
        return 0.0  # pragma: no cover


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.stemmer._snowball_norwegian.

Snowball Norwegian stemmer
"""

from unicodedata import normalize

from deprecation import deprecated

from ._snowball import _Snowball
from .. import __version__

__all__ = ['SnowballNorwegian', 'sb_norwegian']


class SnowballNorwegian(_Snowball):
    """Snowball Norwegian stemmer.

    The Snowball Norwegian stemmer is defined at:
    http://snowball.tartarus.org/algorithms/norwegian/stemmer.html

    .. versionadded:: 0.3.6
    """

    _vowels = {'a', 'e', 'i', 'o', 'u', 'y', 'å', 'æ', 'ø'}
    _s_endings = {
        'b',
        'c',
        'd',
        'f',
        'g',
        'h',
        'j',
        'l',
        'm',
        'n',
        'o',
        'p',
        'r',
        't',
        'v',
        'y',
        'z',
    }

    def stem(self, word):
        """Return Snowball Norwegian stem.

        Parameters
        ----------
        word : str
            The word to stem

        Returns
        -------
        str
            Word stem

        Examples
        --------
        >>> stmr = SnowballNorwegian()
        >>> stmr.stem('lese')
        'les'
        >>> stmr.stem('suspensjon')
        'suspensjon'
        >>> stmr.stem('sikkerhet')
        'sikker'


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        # lowercase, normalize, and compose
        word = normalize('NFC', word.lower())

        r1_start = min(max(3, self._sb_r1(word)), len(word))

        # Step 1
        _r1 = word[r1_start:]
        if _r1[-7:] == 'hetenes':
            word = word[:-7]
        elif _r1[-6:] in {'hetene', 'hetens'}:
            word = word[:-6]
        elif _r1[-5:] in {'heten', 'heter', 'endes'}:
            word = word[:-5]
        elif _r1[-4:] in {'ande', 'ende', 'edes', 'enes', 'erte'}:
            if word[-4:] == 'erte':
                word = word[:-2]
            else:
                word = word[:-4]
        elif _r1[-3:] in {
            'ede',
            'ane',
            'ene',
            'ens',
            'ers',
            'ets',
            'het',
            'ast',
            'ert',
        }:
            if word[-3:] == 'ert':
                word = word[:-1]
            else:
                word = word[:-3]
        elif _r1[-2:] in {'en', 'ar', 'er', 'as', 'es', 'et'}:
            word = word[:-2]
        elif _r1[-1:] in {'a', 'e'}:
            word = word[:-1]
        elif _r1[-1:] == 's':
            if (len(word) > 1 and word[-2] in self._s_endings) or (
                len(word) > 2
                and word[-2] == 'k'
                and word[-3] not in self._vowels
            ):
                word = word[:-1]

        # Step 2
        if word[r1_start:][-2:] in {'dt', 'vt'}:
            word = word[:-1]

        # Step 3
        _r1 = word[r1_start:]
        if _r1[-7:] == 'hetslov':
            word = word[:-7]
        elif _r1[-4:] in {'eleg', 'elig', 'elov', 'slov'}:
            word = word[:-4]
        elif _r1[-3:] in {'leg', 'eig', 'lig', 'els', 'lov'}:
            word = word[:-3]
        elif _r1[-2:] == 'ig':
            word = word[:-2]

        return word


@deprecated(
    deprecated_in='0.4.0',
    removed_in='0.6.0',
    current_version=__version__,
    details='Use the SnowballNorwegian.stem method instead.',
)
def sb_norwegian(word):
    """Return Snowball Norwegian stem.

    This is a wrapper for :py:meth:`SnowballNorwegian.stem`.

    Parameters
    ----------
    word : str
        The word to stem

    Returns
    -------
    str
        Word stem

    Examples
    --------
    >>> sb_norwegian('lese')
    'les'
    >>> sb_norwegian('suspensjon')
    'suspensjon'
    >>> sb_norwegian('sikkerhet')
    'sikker'

    .. versionadded:: 0.1.0

    """
    return SnowballNorwegian().stem(word)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._legalipy.

LegaliPy tokenizer class
"""

from ._tokenizer import _Tokenizer

try:
    from syllabipy.legalipy import LegaliPy
    from syllabipy.legalipy import getOnsets as gen_onsets  # noqa: N813
except ImportError:  # pragma: no cover
    # If the system lacks the SyllabiPy library, that's fine, but SyllabiPy
    # tokenization won't be supported.
    gen_onsets = None
    LegaliPy = None


class LegaliPyTokenizer(_Tokenizer):
    """LegaliPy tokenizer.

    .. versionadded:: 0.4.0
    """

    def __init__(self, scaler=None):
        """Initialize Tokenizer.

        Parameters
        ----------
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - 'length' : Each token has weight equal to its length.
                - 'length-log' : Each token has weight equal to the log of its
                   length + 1.
                - 'length-exp' : Each token has weight equal to e raised to its
                   length.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.


        .. versionadded:: 0.4.0

        """
        if LegaliPy is None:
            raise TypeError(  # pragma: no cover
                'LegaliPy tokenizer requires installation of SyllabiPy'
                + ' package.'
            )

        super(LegaliPyTokenizer, self).__init__(scaler)

        self._onsets = ['']

    def train_onsets(self, text, threshold=0.0002, clean=True, append=False):
        """Train the onsets on a text.

        Parameters
        ----------
        text : str
            The text on which to train
        threshold : float
            Threshold proportion above which to include onset into onset list
        clean : bool
            If True, the text is stripped of numerals and punctuation
        append : bool
            If True, the current onset list is extended


        .. versionadded:: 0.4.0

        """
        new_onsets = gen_onsets(text, threshold, clean)
        if append:
            self._onsets = list(set(self._onsets + new_onsets))
        else:
            self._onsets = new_onsets

    def tokenize(self, string, ipa=False):
        """Tokenize the term and store it.

        The tokenized term is stored as an ordered list and as a Counter
        object.

        Parameters
        ----------
        string : str
            The string to tokenize
        ipa : bool
            If True, indicates that the string is in IPA

        Examples
        --------
        >>> LegaliPyTokenizer().tokenize('seven-twelfths')
        LegaliPyTokenizer({'s': 1, 'ev': 1, 'en-tw': 1, 'elfths': 1})

        >>> LegaliPyTokenizer().tokenize('character')
        LegaliPyTokenizer({'ch': 1, 'ar': 1, 'act': 1, 'er': 1})

        .. versionadded:: 0.4.0

        """
        self._string = string

        self._ordered_tokens = []
        for word in string.split():
            self._ordered_tokens += LegaliPy(word, self._onsets)
        if not self._ordered_tokens:
            self._ordered_tokens = [self._string]

        super(LegaliPyTokenizer, self).tokenize()
        return self


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._q_skipgrams.

Q-Skipgrams multi-set class
"""

from collections.abc import Iterable
from itertools import combinations

from ._tokenizer import _Tokenizer

__all__ = ['QSkipgrams']


class QSkipgrams(_Tokenizer):
    """A q-skipgram class, which functions like a bag/multiset.

    A q-gram is here defined as all sequences of q characters. Q-grams are also
    known as k-grams and n-grams, but the term n-gram more typically refers to
    sequences of whitespace-delimited words in a string, where q-gram refers
    to sequences of characters in a word or string.

    .. versionadded:: 0.4.0
    """

    def __init__(self, qval=2, start_stop='$#', scaler=None, ssk_lambda=0.9):
        """Initialize QSkipgrams.

        Parameters
        ----------
        qval : int or Iterable
            The q-gram length (defaults to 2), can be an integer, range object,
            or list
        start_stop : str
            A string of length >= 0 indicating start & stop symbols.
            If the string is '', q-grams will be calculated without start &
            stop symbols appended to each end.
            Otherwise, the first character of start_stop will pad the
            beginning of the string and the last character of start_stop
            will pad the end of the string before q-grams are calculated.
            (In the case that start_stop is only 1 character long, the same
            symbol will be used for both.)
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - 'length' : Each token has weight equal to its length.
                - 'length-log' : Each token has weight equal to the log of its
                   length + 1.
                - 'length-exp' : Each token has weight equal to e raised to its
                   length.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.
                - 'SSK' : Applies weighting according to the substring kernel
                  rules of :cite:`Lodhi:2002`.
        ssk_lambda : float or Iterable
            A value in the range (0.0, 1.0) used for discouting gaps between
            characters according to the method described in :cite:`Lodhi:2002`.
            To supply multiple values of lambda, provide an Iterable of numeric
            values, such as (0.5, 0.05) or np.arange(0.05, 0.5, 0.05)

        Raises
        ------
        ValueError
            Use WhitespaceTokenizer instead of qval=0.

        Examples
        --------
        >>> QSkipgrams().tokenize('AATTAT')
        QSkipgrams({'AT': 7, '$A': 3, '$T': 3, 'AA': 3, 'A#': 3, 'TT': 3,
        'T#': 3, 'TA': 2, '$#': 1})

        >>> QSkipgrams(qval=1, start_stop='').tokenize('AATTAT')
        QSkipgrams({'A': 3, 'T': 3})

        >>> QSkipgrams(qval=3, start_stop='').tokenize('AATTAT')
        QSkipgrams({'ATT': 6, 'AAT': 5, 'ATA': 4, 'TAT': 2, 'AAA': 1,
        'TTA': 1, 'TTT': 1})

        >>> QSkipgrams(start_stop='').tokenize('ABCD')
        QSkipgrams({'AB': 1, 'AC': 1, 'AD': 1, 'BC': 1, 'BD': 1, 'CD': 1})

        >>> QSkipgrams().tokenize('Colin')
        QSkipgrams({'$C': 1, '$o': 1, '$l': 1, '$i': 1, '$n': 1, '$#': 1,
        'Co': 1, 'Cl': 1, 'Ci': 1, 'Cn': 1, 'C#': 1, 'ol': 1, 'oi': 1, 'on': 1,
        'o#': 1, 'li': 1, 'ln': 1, 'l#': 1, 'in': 1, 'i#': 1, 'n#': 1})

        >>> QSkipgrams(qval=3).tokenize('AACTAGAAC')
        QSkipgrams({'$AA': 20, '$A#': 20, 'AA#': 20, '$AC': 14, 'AC#': 14,
        'AAC': 11, 'AAA': 10, '$C#': 8, '$AG': 6, '$CA': 6, '$TA': 6, 'ACA': 6,
        'ATA': 6, 'AGA': 6, 'AG#': 6, 'CA#': 6, 'TA#': 6, '$$A': 5, 'A##': 5,
        '$AT': 4, '$T#': 4, '$GA': 4, '$G#': 4, 'AT#': 4, 'GA#': 4, 'AAG': 3,
        'AGC': 3, 'CTA': 3, 'CAA': 3, 'CAC': 3, 'TAA': 3, 'TAC': 3, '$$C': 2,
        '$$#': 2, '$CT': 2, '$CG': 2, '$CC': 2, '$TG': 2, '$TC': 2, '$GC': 2,
        '$##': 2, 'ACT': 2, 'ACG': 2, 'ACC': 2, 'ATG': 2, 'ATC': 2, 'CT#': 2,
        'CGA': 2, 'CG#': 2, 'CC#': 2, 'C##': 2, 'TGA': 2, 'TG#': 2, 'TC#': 2,
        'GAC': 2, 'GC#': 2, '$$T': 1, '$$G': 1, 'AAT': 1, 'CTG': 1, 'CTC': 1,
        'CAG': 1, 'CGC': 1, 'TAG': 1, 'TGC': 1, 'T##': 1, 'GAA': 1, 'G##': 1})

        QSkipgrams may also be used to produce weights in accordance with the
        substring kernel rules of :cite:`Lodhi:2002` by passing the scaler
        value ``'SSK'``:

        >>> QSkipgrams(scaler='SSK').tokenize('AACTAGAAC')
        QSkipgrams({'AA': 6.170192010000001, 'AC': 4.486377699,
        '$A': 2.8883286990000006, 'A#': 2.6526399291000002, 'TA': 2.05659,
        'AG': 1.931931, 'CA': 1.850931, 'GA': 1.5390000000000001, 'AT': 1.3851,
        'C#': 1.2404672100000003, '$C': 1.0047784401000002, 'CT': 0.81,
        'TG': 0.7290000000000001, 'CG': 0.6561, 'GC': 0.6561,
        '$T': 0.5904900000000001, 'G#': 0.5904900000000001, 'TC': 0.531441,
        '$G': 0.4782969000000001, 'CC': 0.4782969000000001,
        'T#': 0.4782969000000001, '$#': 0.31381059609000006})

        .. versionadded:: 0.4.0

        """
        super(QSkipgrams, self).__init__(scaler)

        # Save parameters
        self.qval = qval
        self.start_stop = start_stop
        if qval == 1:
            self.start_stop = ''

        self._string_ss = self._string
        if isinstance(ssk_lambda, float):
            self._lambda = (ssk_lambda,)
        else:
            self._lambda = tuple(ssk_lambda)

    def tokenize(self, string):
        """Tokenize the term and store it.

        The tokenized term is stored as an ordered list and as a Counter
        object.

        Parameters
        ----------
        string : str
            The string to tokenize


        .. versionadded:: 0.4.0

        """
        self._string = string
        self._ordered_tokens = []
        self._ordered_weights = []

        if not isinstance(self.qval, Iterable):
            self.qval = (self.qval,)

        for qval_i in self.qval:
            if qval_i < 1:
                continue

            if self.start_stop and self._string:
                string = (
                    self.start_stop[0] * (qval_i - 1)
                    + self._string
                    + self.start_stop[-1] * (qval_i - 1)
                )
            else:
                string = self._string

            if len(string) < qval_i:
                continue

            # Having appended start & stop symbols (or not), save the
            # result, but only for the longest valid qval_i
            if len(string) > len(self._string_ss):
                self._string_ss = string

            combs = list(combinations(enumerate(string), qval_i))
            self._ordered_tokens += [''.join(l[1] for l in t) for t in combs]

            if self._scaler == 'SSK':
                self._ordered_weights += [
                    sum(
                        l ** (t[-1][0] - t[0][0] + len(t) - 1)
                        for l in self._lambda
                    )
                    for t in combs
                ]
            else:
                self._ordered_weights += [1] * len(combs)

        super(QSkipgrams, self).tokenize()
        return self


if __name__ == '__main__':
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.fingerprint._lacss.

L.A. County Sheriff's System fingerprint
"""

from ._fingerprint import _Fingerprint

__all__ = ['LACSS']


class LACSS(_Fingerprint):
    """L.A. County Sheriff's System fingerprint.

    Based on the description from :cite:`Taft:1970`.

    .. versionadded:: 0.4.1
    """

    _vowels = set('AEIOUYWH')

    _t1 = {_[0]: _[1] for _ in zip('RNLTSCMDKAGBPFVZXJQ', range(1, 20))}
    _t1.update({_: 0 for _ in _vowels})

    _t2 = {_[0]: _[1] for _ in zip('ABCDEFGHIJKLMNOPQRSTUVWXYZ', range(1, 27))}

    def fingerprint(self, word):
        """Return the LACSS coding.

        Parameters
        ----------
        word : str
            The word to fingerprint

        Returns
        -------
        int
            The L.A. County Sheriff's System fingerprint

        Examples
        --------
        >>> cf = LACSS()
        >>> cf.fingerprint('hat')
        '4911211'
        >>> cf.fingerprint('niall')
        '6488374'
        >>> cf.fingerprint('colin')
        '3015957'
        >>> cf.fingerprint('atcg')
        '1772371'
        >>> cf.fingerprint('entreatment')
        '3882324'


        .. versionadded:: 0.4.1

        """
        # uppercase
        word = word.upper()

        # remove vowels
        word = word[:1] + ''.join(_ for _ in word[1:] if _ not in 'AEIOUWHY')
        word += 12 * 'A'

        # step 1
        code = 0
        i = 0
        while (not code) and (i < len(word)):
            if word[i] in self._t2:
                code = self._t2[word[i]] * 10
            i += 1

        letters = 11
        while letters and i < len(word):
            if word[i] in self._t1:
                code *= 10
                code += self._t1[word[i]]
                letters -= 1
            i += 1

        code *= 3
        code = str(int(code ** 0.5))

        return code


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.phonetic._phonem.

Phonem
"""

from unicodedata import normalize as unicode_normalize

from deprecation import deprecated

from ._phonetic import _Phonetic
from .. import __version__

__all__ = ['Phonem', 'phonem']


class Phonem(_Phonetic):
    """Phonem.

    Phonem is defined in :cite:`Wilde:1988`.

    This version is based on the Perl implementation documented at
    :cite:`Wilz:2005`.
    It includes some enhancements presented in the Java port at
    :cite:`dcm4che:2011`.

    Phonem is intended chiefly for German names/words.

    .. versionadded:: 0.3.6
    """

    _substitutions = (
        ('SC', 'C'),
        ('SZ', 'C'),
        ('CZ', 'C'),
        ('TZ', 'C'),
        ('TS', 'C'),
        ('KS', 'X'),
        ('PF', 'V'),
        ('QU', 'KW'),
        ('PH', 'V'),
        ('UE', 'Y'),
        ('AE', 'E'),
        ('OE', 'Ö'),
        ('EI', 'AY'),
        ('EY', 'AY'),
        ('EU', 'OY'),
        ('AU', 'A§'),
        ('OU', '§'),
    )

    _trans = dict(
        zip(
            (ord(_) for _ in 'ZKGQÇÑßFWPTÁÀÂÃÅÄÆÉÈÊËIJÌÍÎÏÜÝ§ÚÙÛÔÒÓÕØ'),
            'CCCCCNSVVBDAAAAAEEEEEEYYYYYYYYUUUUOOOOÖ',
        )
    )

    _uc_set = set('ABCDLMNORSUVWXYÖ')

    def encode(self, word):
        """Return the Phonem code for a word.

        Parameters
        ----------
        word : str
        The word to transform

        Returns
        -------
        str
            The Phonem value

        Examples
        --------
        >>> pe = Phonem()
        >>> pe.encode('Christopher')
        'CRYSDOVR'
        >>> pe.encode('Niall')
        'NYAL'
        >>> pe.encode('Smith')
        'SMYD'
        >>> pe.encode('Schmidt')
        'CMYD'


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        word = unicode_normalize('NFC', word.upper())
        for i, j in self._substitutions:
            word = word.replace(i, j)
        word = word.translate(self._trans)

        return ''.join(
            c
            for c in self._delete_consecutive_repeats(word)
            if c in self._uc_set
        )


@deprecated(
    deprecated_in='0.4.0',
    removed_in='0.6.0',
    current_version=__version__,
    details='Use the Phonem.encode method instead.',
)
def phonem(word):
    """Return the Phonem code for a word.

    This is a wrapper for :py:meth:`Phonem.encode`.

    Parameters
    ----------
    word : str
        The word to transform

    Returns
    -------
    str
        The Phonem value

    Examples
    --------
    >>> phonem('Christopher')
    'CRYSDOVR'
    >>> phonem('Niall')
    'NYAL'
    >>> phonem('Smith')
    'SMYD'
    >>> phonem('Schmidt')
    'CMYD'

    .. versionadded:: 0.1.0

    """
    return Phonem().encode(word)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.phonetic._alpha_sis.

IBM's Alpha Search Inquiry System coding
"""

from unicodedata import normalize as unicode_normalize

from deprecation import deprecated

from ._phonetic import _Phonetic
from .. import __version__

__all__ = ['AlphaSIS', 'alpha_sis']


class AlphaSIS(_Phonetic):
    """Alpha-SIS.

    The Alpha Search Inquiry System code is defined in :cite:`IBM:1973`.
    This implementation is based on the description in :cite:`Moore:1977`.

    .. versionadded:: 0.3.6
    """

    _alpha_sis_initials = {
        'GF': '08',
        'GM': '03',
        'GN': '02',
        'KN': '02',
        'PF': '08',
        'PN': '02',
        'PS': '00',
        'WR': '04',
        'A': '1',
        'E': '1',
        'H': '2',
        'I': '1',
        'J': '3',
        'O': '1',
        'U': '1',
        'W': '4',
        'Y': '5',
    }
    _alpha_sis_initials_order = (
        'GF',
        'GM',
        'GN',
        'KN',
        'PF',
        'PN',
        'PS',
        'WR',
        'A',
        'E',
        'H',
        'I',
        'J',
        'O',
        'U',
        'W',
        'Y',
    )
    _alpha_sis_basic = {
        'SCH': '6',
        'CZ': ('70', '6', '0'),
        'CH': ('6', '70', '0'),
        'CK': ('7', '6'),
        'DS': ('0', '10'),
        'DZ': ('0', '10'),
        'TS': ('0', '10'),
        'TZ': ('0', '10'),
        'CI': '0',
        'CY': '0',
        'CE': '0',
        'SH': '6',
        'DG': '7',
        'PH': '8',
        'C': ('7', '6'),
        'K': ('7', '6'),
        'Z': '0',
        'S': '0',
        'D': '1',
        'T': '1',
        'N': '2',
        'M': '3',
        'R': '4',
        'L': '5',
        'J': '6',
        'G': '7',
        'Q': '7',
        'X': '7',
        'F': '8',
        'V': '8',
        'B': '9',
        'P': '9',
    }
    _alpha_sis_basic_order = (
        'SCH',
        'CZ',
        'CH',
        'CK',
        'DS',
        'DZ',
        'TS',
        'TZ',
        'CI',
        'CY',
        'CE',
        'SH',
        'DG',
        'PH',
        'C',
        'K',
        'Z',
        'S',
        'D',
        'T',
        'N',
        'M',
        'R',
        'L',
        'J',
        'C',
        'G',
        'K',
        'Q',
        'X',
        'F',
        'V',
        'B',
        'P',
    )

    _alphabetic_initials = dict(zip((ord(_) for _ in '012345'), ' AHJWY'))
    _alphabetic_non_initials = dict(
        zip((ord(_) for _ in '0123456789'), 'STNMRLJKFP')
    )

    def __init__(self, max_length=14):
        """Initialize AlphaSIS instance.

        Parameters
        ----------
        max_length : int
            The length of the code returned (defaults to 14)


        .. versionadded:: 0.4.0

        """
        # Clamp max_length to [4, 64]
        if max_length != -1:
            self._max_length = min(max(4, max_length), 64)
        else:
            self._max_length = 64

    def encode_alpha(self, word):
        """Return the alphabetic Alpha-SIS code for a word.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        tuple
            The alphabetic Alpha-SIS value

        Examples
        --------
        >>> pe = AlphaSIS()
        >>> pe.encode_alpha('Christopher')
        ('JRSTFR', 'KSRSTFR', 'RSTFR')
        >>> pe.encode_alpha('Niall')
        ('NL',)
        >>> pe.encode_alpha('Smith')
        ('MT',)
        >>> pe.encode_alpha('Schmidt')
        ('JMT',)


        .. versionadded:: 0.4.0

        """
        codes = self.encode(word)
        alphas = [
            code[0].translate(self._alphabetic_initials).strip()
            + code[1:].translate(self._alphabetic_non_initials).rstrip('S')
            for code in codes
        ]

        return tuple(alphas)

    def encode(self, word):
        """Return the IBM Alpha Search Inquiry System code for a word.

        A collection is necessary as the return type since there can be
        multiple values for a single word. But the collection must be ordered
        since the first value is the primary coding.

        Parameters
        ----------
        word : str
            The word to transform

        Returns
        -------
        tuple
            The Alpha-SIS value

        Examples
        --------
        >>> pe = AlphaSIS()
        >>> pe.encode('Christopher')
        ('06401840000000', '07040184000000', '04018400000000')
        >>> pe.encode('Niall')
        ('02500000000000',)
        >>> pe.encode('Smith')
        ('03100000000000',)
        >>> pe.encode('Schmidt')
        ('06310000000000',)


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        alpha = ['']
        pos = 0
        word = unicode_normalize('NFKD', word.upper())
        word = ''.join(c for c in word if c in self._uc_set)

        # Do special processing for initial substrings
        for k in self._alpha_sis_initials_order:
            if word.startswith(k):
                alpha[0] += self._alpha_sis_initials[k]
                pos += len(k)
                break

        # Add a '0' if alpha is still empty
        if not alpha[0]:
            alpha[0] += '0'

        # Whether or not any special initial codes were encoded, iterate
        # through the length of the word in the main encoding loop
        while pos < len(word):
            orig_pos = pos
            for k in self._alpha_sis_basic_order:
                if word[pos:].startswith(k):
                    if isinstance(self._alpha_sis_basic[k], tuple):
                        newalpha = []
                        for i in range(len(self._alpha_sis_basic[k])):
                            newalpha += [
                                _ + self._alpha_sis_basic[k][i] for _ in alpha
                            ]
                        alpha = newalpha
                    else:
                        alpha = [_ + self._alpha_sis_basic[k] for _ in alpha]
                    pos += len(k)
                    break
            if pos == orig_pos:
                alpha = [_ + '_' for _ in alpha]
                pos += 1

        # Trim doublets and placeholders
        for i in range(len(alpha)):
            pos = 1
            while pos < len(alpha[i]):
                if alpha[i][pos] == alpha[i][pos - 1]:
                    alpha[i] = alpha[i][:pos] + alpha[i][pos + 1 :]
                pos += 1
        alpha = (_.replace('_', '') for _ in alpha)

        # Trim codes and return tuple
        alpha = (
            (_ + ('0' * self._max_length))[: self._max_length] for _ in alpha
        )
        return tuple(alpha)


@deprecated(
    deprecated_in='0.4.0',
    removed_in='0.6.0',
    current_version=__version__,
    details='Use the AlphaSIS.encode method instead.',
)
def alpha_sis(word, max_length=14):
    """Return the IBM Alpha Search Inquiry System code for a word.

    This is a wrapper for :py:meth:`AlphaSIS.encode`.

    Parameters
    ----------
    word : str
        The word to transform
    max_length : int
        The length of the code returned (defaults to 14)

    Returns
    -------
    tuple
        The Alpha-SIS value

    Examples
    --------
    >>> alpha_sis('Christopher')
    ('06401840000000', '07040184000000', '04018400000000')
    >>> alpha_sis('Niall')
    ('02500000000000',)
    >>> alpha_sis('Smith')
    ('03100000000000',)
    >>> alpha_sis('Schmidt')
    ('06310000000000',)

    .. versionadded:: 0.1.0

    """
    return AlphaSIS(max_length).encode(word)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._fuzzywuzzy_token_sort.

FuzzyWuzzy Token Sort similarity
"""

from difflib import SequenceMatcher

from ._token_distance import _TokenDistance
from ..tokenizer import RegexpTokenizer

__all__ = ['FuzzyWuzzyTokenSort']


class FuzzyWuzzyTokenSort(_TokenDistance):
    r"""FuzzyWuzzy Token Sort similarity.

    This follows the FuzzyWuzzy Token Sort similarity algorithm
    :cite:`Cohen:2011`. Rather than returning an integer in the range [0, 100],
    as demonstrated in the blog post, this implementation returns a float in
    the range [0.0, 1.0].

    .. versionadded:: 0.4.0
    """

    def __init__(self, tokenizer=None, **kwargs):
        """Initialize FuzzyWuzzyTokenSort instance.

        Parameters
        ----------
        tokenizer : _Tokenizer
            A tokenizer instance from the :py:mod:`abydos.tokenizer` package.
            By default, the regexp tokenizer is employed, matching only
            letters.
        **kwargs
            Arbitrary keyword arguments

        Other Parameters
        ----------------
        qval : int
            The length of each q-gram. Using this parameter and tokenizer=None
            will cause the instance to use the QGram tokenizer with this
            q value.


        .. versionadded:: 0.4.0

        """
        if tokenizer is None:
            tokenizer = RegexpTokenizer()
        super(FuzzyWuzzyTokenSort, self).__init__(
            tokenizer=tokenizer, **kwargs
        )

    def sim(self, src, tar):
        """Return the FuzzyWuzzy Token Sort similarity of two strings.

        Parameters
        ----------
        src : str
            Source string (or QGrams/Counter objects) for comparison
        tar : str
            Target string (or QGrams/Counter objects) for comparison

        Returns
        -------
        float
            FuzzyWuzzy Token Sort similarity

        Examples
        --------
        >>> cmp = FuzzyWuzzyTokenSort()
        >>> cmp.sim('cat', 'hat')
        0.6666666666666666
        >>> cmp.sim('Niall', 'Neil')
        0.6666666666666666
        >>> cmp.sim('aluminum', 'Catalan')
        0.4
        >>> cmp.sim('ATCG', 'TAGC')
        0.5


        .. versionadded:: 0.4.0

        """
        src = ' '.join(
            sorted(self.params['tokenizer'].tokenize(src).get_list())
        )
        tar = ' '.join(
            sorted(self.params['tokenizer'].tokenize(tar).get_list())
        )

        return SequenceMatcher(None, src, tar).ratio()


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._character.

Character tokenizer
"""

from ._tokenizer import _Tokenizer

__all__ = ['CharacterTokenizer']


class CharacterTokenizer(_Tokenizer):
    """A character tokenizer.

    .. versionadded:: 0.4.0
    """

    def __init__(self, scaler=None):
        """Initialize tokenizer.

        Parameters
        ----------
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.


        .. versionadded:: 0.4.0

        """
        super(CharacterTokenizer, self).__init__(scaler)

    def tokenize(self, string):
        """Tokenize the term and store it.

        The tokenized term is stored as an ordered list and as a Counter
        object.

        Parameters
        ----------
        string : str
            The string to tokenize

        Examples
        --------
        >>> CharacterTokenizer().tokenize('AACTAGAAC')
        CharacterTokenizer({'A': 5, 'C': 2, 'T': 1, 'G': 1})

        .. versionadded:: 0.4.0

        """
        self._string = string
        self._ordered_tokens = list(string)

        super(CharacterTokenizer, self).tokenize()
        return self


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._eudex.

eudex distance functions
"""

from types import GeneratorType

from deprecation import deprecated

from ._distance import _Distance
from .. import __version__
from ..phonetic import eudex

__all__ = ['Eudex', 'dist_eudex', 'eudex_hamming', 'sim_eudex']


class Eudex(_Distance):
    """Distance between the Eudex hashes of two terms.

    Cf. :cite:`Ticki:2016`.

    .. versionadded:: 0.3.6
    """

    @staticmethod
    def gen_fibonacci():
        """Yield the next Fibonacci number.

        Based on https://www.python-course.eu/generators.php
        Starts at Fibonacci number 3 (the second 1)

        Yields
        ------
        int
            The next Fibonacci number


        .. versionadded:: 0.3.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        num_a, num_b = 1, 2
        while True:
            yield num_a
            num_a, num_b = num_b, num_a + num_b

    @staticmethod
    def gen_exponential(base=2):
        """Yield the next value in an exponential series of the base.

        Starts at base**0

        Parameters
        ----------
        base : int
            The base to exponentiate

        Yields
        ------
        int
            The next power of `base`


        .. versionadded:: 0.3.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        exp = 0
        while True:
            yield base ** exp
            exp += 1

    def __init__(self, weights='exponential', max_length=8, **kwargs):
        """Initialize Eudex instance.

        Parameters
        ----------
        weights : str, iterable, or generator function
            The weights or weights generator function

                - If set to ``None``, a simple Hamming distance is calculated.
                - If set to ``exponential``, weight decays by powers of 2, as
                  proposed in the eudex specification:
                  https://github.com/ticki/eudex.
                - If set to ``fibonacci``, weight decays through the Fibonacci
                  series, as in the eudex reference implementation.
                - If set to a callable function, this assumes it creates a
                  generator and the generator is used to populate a series of
                  weights.
                - If set to an iterable, the iterable's values should be
                  integers and will be used as the weights.

            In all cases, the weights should be ordered or generated from least
            significant to most significant, so larger values should generally
            come first.

        max_length : int
            The number of characters to encode as a eudex hash
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        super(Eudex, self).__init__(**kwargs)
        self._weights = weights
        self._max_length = max_length

    def dist_abs(self, src, tar, normalized=False):
        """Calculate the distance between the Eudex hashes of two terms.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison
        normalized : bool
            Normalizes to [0, 1] if True

        Returns
        -------
        int
            The Eudex Hamming distance

        Examples
        --------
        >>> cmp = Eudex()
        >>> cmp.dist_abs('cat', 'hat')
        128
        >>> cmp.dist_abs('Niall', 'Neil')
        2
        >>> cmp.dist_abs('Colin', 'Cuilen')
        10
        >>> cmp.dist_abs('ATCG', 'TAGC')
        403

        >>> cmp = Eudex(weights='fibonacci')
        >>> cmp.dist_abs('cat', 'hat')
        34
        >>> cmp.dist_abs('Niall', 'Neil')
        2
        >>> cmp.dist_abs('Colin', 'Cuilen')
        7
        >>> cmp.dist_abs('ATCG', 'TAGC')
        117

        >>> cmp = Eudex(weights=None)
        >>> cmp.dist_abs('cat', 'hat')
        1
        >>> cmp.dist_abs('Niall', 'Neil')
        1
        >>> cmp.dist_abs('Colin', 'Cuilen')
        2
        >>> cmp.dist_abs('ATCG', 'TAGC')
        9

        >>> # Using the OEIS A000142:
        >>> cmp = Eudex(weights=[1, 1, 2, 6, 24, 120, 720, 5040])
        >>> cmp.dist_abs('cat', 'hat')
        5040
        >>> cmp.dist_abs('Niall', 'Neil')
        1
        >>> cmp.dist_abs('Colin', 'Cuilen')
        7
        >>> cmp.dist_abs('ATCG', 'TAGC')
        15130


        .. versionadded:: 0.3.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """

        # Calculate the eudex hashes and XOR them
        xored = eudex(src, max_length=self._max_length) ^ eudex(
            tar, max_length=self._max_length
        )

        # Simple hamming distance (all bits are equal)
        if not self._weights:
            binary = bin(xored)
            distance = binary.count('1')
            if normalized:
                return distance / (len(binary) - 2)
            return distance

        # If weights is a function, it should create a generator,
        # which we now use to populate a list
        if callable(self._weights):
            weights = self._weights()
        elif self._weights == 'exponential':
            weights = Eudex.gen_exponential()
        elif self._weights == 'fibonacci':
            weights = Eudex.gen_fibonacci()
        elif hasattr(self._weights, '__iter__') and not isinstance(
            self._weights, str
        ):
            weights = self._weights[::-1]
        else:
            raise ValueError('Unrecognized weights value or type.')

        if isinstance(weights, GeneratorType):
            weights = [next(weights) for _ in range(self._max_length)][::-1]

        # Sum the weighted hamming distance
        distance = 0
        max_distance = 0
        while (xored or normalized) and weights:
            max_distance += 8 * weights[-1]
            distance += bin(xored & 0xFF).count('1') * weights.pop()
            xored >>= 8

        if normalized:
            distance /= max_distance

        return distance

    def dist(self, src, tar):
        """Return normalized distance between the Eudex hashes of two terms.

        This is Eudex distance normalized to [0, 1].

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        int
            The normalized Eudex Hamming distance

        Examples
        --------
        >>> cmp = Eudex()
        >>> round(cmp.dist('cat', 'hat'), 12)
        0.062745098039
        >>> round(cmp.dist('Niall', 'Neil'), 12)
        0.000980392157
        >>> round(cmp.dist('Colin', 'Cuilen'), 12)
        0.004901960784
        >>> round(cmp.dist('ATCG', 'TAGC'), 12)
        0.197549019608


        .. versionadded:: 0.3.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        return self.dist_abs(src, tar, True)


@deprecated(
    deprecated_in='0.4.0',
    removed_in='0.6.0',
    current_version=__version__,
    details='Use the Eudex.dist_abs method instead.',
)
def eudex_hamming(
    src, tar, weights='exponential', max_length=8, normalized=False
):
    """Calculate the Hamming distance between the Eudex hashes of two terms.

    This is a wrapper for :py:meth:`Eudex.eudex_hamming`.

    Parameters
    ----------
    src : str
        Source string for comparison
    tar : str
        Target string for comparison
    weights : str, iterable, or generator function
        The weights or weights generator function
    max_length : int
        The number of characters to encode as a eudex hash
    normalized : bool
        Normalizes to [0, 1] if True

    Returns
    -------
    int
        The Eudex Hamming distance

    Examples
    --------
    >>> eudex_hamming('cat', 'hat')
    128
    >>> eudex_hamming('Niall', 'Neil')
    2
    >>> eudex_hamming('Colin', 'Cuilen')
    10
    >>> eudex_hamming('ATCG', 'TAGC')
    403

    >>> eudex_hamming('cat', 'hat', weights='fibonacci')
    34
    >>> eudex_hamming('Niall', 'Neil', weights='fibonacci')
    2
    >>> eudex_hamming('Colin', 'Cuilen', weights='fibonacci')
    7
    >>> eudex_hamming('ATCG', 'TAGC', weights='fibonacci')
    117

    >>> eudex_hamming('cat', 'hat', weights=None)
    1
    >>> eudex_hamming('Niall', 'Neil', weights=None)
    1
    >>> eudex_hamming('Colin', 'Cuilen', weights=None)
    2
    >>> eudex_hamming('ATCG', 'TAGC', weights=None)
    9

    >>> # Using the OEIS A000142:
    >>> eudex_hamming('cat', 'hat', [1, 1, 2, 6, 24, 120, 720, 5040])
    5040
    >>> eudex_hamming('Niall', 'Neil', [1, 1, 2, 6, 24, 120, 720, 5040])
    1
    >>> eudex_hamming('Colin', 'Cuilen', [1, 1, 2, 6, 24, 120, 720, 5040])
    7
    >>> eudex_hamming('ATCG', 'TAGC', [1, 1, 2, 6, 24, 120, 720, 5040])
    15130

    .. versionadded:: 0.3.0

    """
    return Eudex(weights, max_length).dist_abs(src, tar, normalized)


@deprecated(
    deprecated_in='0.4.0',
    removed_in='0.6.0',
    current_version=__version__,
    details='Use the Eudex.dist method instead.',
)
def dist_eudex(src, tar, weights='exponential', max_length=8):
    """Return normalized Hamming distance between Eudex hashes of two terms.

    This is a wrapper for :py:meth:`Eudex.dist`.

    Parameters
    ----------
    src : str
        Source string for comparison
    tar : str
        Target string for comparison
    weights : str, iterable, or generator function
        The weights or weights generator function
    max_length : int
        The number of characters to encode as a eudex hash

    Returns
    -------
    int
        The normalized Eudex Hamming distance

    Examples
    --------
    >>> round(dist_eudex('cat', 'hat'), 12)
    0.062745098039
    >>> round(dist_eudex('Niall', 'Neil'), 12)
    0.000980392157
    >>> round(dist_eudex('Colin', 'Cuilen'), 12)
    0.004901960784
    >>> round(dist_eudex('ATCG', 'TAGC'), 12)
    0.197549019608

    .. versionadded:: 0.3.0

    """
    return Eudex(weights, max_length).dist(src, tar)


@deprecated(
    deprecated_in='0.4.0',
    removed_in='0.6.0',
    current_version=__version__,
    details='Use the Eudex.sim method instead.',
)
def sim_eudex(src, tar, weights='exponential', max_length=8):
    """Return normalized Hamming similarity between Eudex hashes of two terms.

    This is a wrapper for :py:meth:`Eudex.sim`.

    Parameters
    ----------
    src : str
        Source string for comparison
    tar : str
        Target string for comparison
    weights : str, iterable, or generator function
        The weights or weights generator function
    max_length : int
        The number of characters to encode as a eudex hash

    Returns
    -------
    int
        The normalized Eudex Hamming similarity

    Examples
    --------
    >>> round(sim_eudex('cat', 'hat'), 12)
    0.937254901961
    >>> round(sim_eudex('Niall', 'Neil'), 12)
    0.999019607843
    >>> round(sim_eudex('Colin', 'Cuilen'), 12)
    0.995098039216
    >>> round(sim_eudex('ATCG', 'TAGC'), 12)
    0.802450980392

    .. versionadded:: 0.3.0

    """
    return Eudex(weights, max_length).sim(src, tar)


if __name__ == '__main__':
    import doctest

    doctest.testmod()
